import { world, MolangVariableMap, StructureSaveMode, system, BlockPermutation, ItemStack } from "@minecraft/server";
import { ModalFormData, ActionFormData } from "@minecraft/server-ui";
import { decrementItemInHand, isCreative } from '../util/utils';
import { allowedMaterialList, farmMap, selectFarmMap } from "./FarmInformation";

world.beforeEvents.worldInitialize.subscribe((e) => {
	e.blockComponentRegistry.registerCustomComponent("if:place_block", new IfPlaceBlock());
	e.blockComponentRegistry.registerCustomComponent("if:set_farm", new IfSetFarm());
	e.itemComponentRegistry.registerCustomComponent("farm:start_config", new IfStartConfig());
	e.itemComponentRegistry.registerCustomComponent("farm:consume_cardboard", new IfCardboardBox());
});
const defaultConfig = {
	blockPercentage: 100,
	showParticles: true,
	volume: 100
};
export const InstaFarmConfig = (() => {
    const storedConfig = world.getDynamicProperty("InstaFarmConfig");
    return storedConfig ? { ...defaultConfig, ...JSON.parse(storedConfig) } : defaultConfig;
})();
//------UI
function AddonInfoUI (player) {
	new ActionFormData()
	.title({ translate: 'farm.title.addon_info' })
	.body({ translate: 'farm.body.addon_info', with: [ '\n' ] })
	.button({ translate: 'farm.button.general_info' }, "textures/farm/button/farm_button")
	.button({ translate: 'farm.button.redstone_table' }, "textures/farm/button/redstone_table")
	.button({ translate: 'farm.button.addon_info_items' }, "textures/farm/items/farm_booster")
	.button({ translate: 'farm.button.material_categories' }, "textures/farm/button/materials")
	.button({ translate: 'farm.button.addon_info_config' }, "textures/farm/items/config")
	.show(player).then(r => {
		if (r.canceled) return;
		[GeneralInfo, RedstoneTableInfoUI, ItemsInfoUI, AllowedMaterialsUI, ConfigUI][r.selection](player);
	});
}
function GeneralInfo (player) {
	new ActionFormData()
	.title({ translate: 'farm.title.general_info' })
	.body({ translate: 'farm.body.general_info', with: [ '\n' ] })
	.button({ translate: 'farm.button.back' }, "textures/farm/button/back")
	.show(player).then(r => {
		if (r.canceled) return;
		[AddonInfoUI][r.selection](player);
	});
}
const farmInfoList = [
	{ buttonTexture: 'button_cactus', name: 'tile.farm:cactus_farm.name', maxLevel: 15, description: 'farm.info.desc.automatic_in_time', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_kelp', name: 'tile.farm:kelp_farm.name', maxLevel: 11, description: 'farm.info.desc.automatic_in_time', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_bamboo', name: 'farm.bamboo_farm_n.name', maxLevel: 10, description: 'farm.info.desc.bamboo_lever', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_bamboo', name: 'farm.bamboo_farm_fm.name', maxLevel: 2, description: 'farm.info.desc.bamboo_flying_machine', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.none', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_sugar_cane', name: 'farm.sugar_cane_farm_n.name', maxLevel: 10, description: 'farm.info.desc.automatic_in_time', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_sugar_cane', name: 'farm.sugar_cane_farm_fm.name', maxLevel: 2, description: 'farm.info.desc.sugar_cane_flying_machine', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_wheat', name: 'tile.farm:wheat_farm.name', maxLevel: 5, description: 'farm.info.desc.crop_farms', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_beetroot', name: 'tile.farm:beetroot_farm.name', maxLevel: 5, description: 'farm.info.desc.crop_farms', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_potato', name: 'tile.farm:potato_farm.name', maxLevel: 5, description: 'farm.info.desc.crop_farms', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_carrot', name: 'tile.farm:carrot_farm.name', maxLevel: 5, description: 'farm.info.desc.crop_farms', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_melon', name: 'tile.farm:melon_farm.name', maxLevel: 15, description: 'farm.info.desc.automatic_in_time', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_pumpkin', name: 'tile.farm:pumpkin_farm.name', maxLevel: 15, description: 'farm.info.desc.automatic_in_time', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_twisting_v', name: 'farm.twisting_vines_farm_n.name', maxLevel: 10, description: 'farm.info.desc.automatic_in_time', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_twisting_v', name: 'farm.twisting_vines_farm_fm.name', maxLevel: 2, description: 'farm.info.desc.twisting_v_flying_machine', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.none', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_warped_roots', name: 'tile.farm:warped_roots_farm.name', maxLevel: 2, description: 'farm.info.desc.nether_roots', requireGamerules: [  ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_weeping_v', name: 'tile.farm:weeping_vines_farm.name', maxLevel: 10, description: 'farm.info.desc.automatic_in_time', requireGamerules: [ 'randomTickSpeed' ], dimensionRestriction: 'farm.info.dimension.none', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_crimson_roots', name: 'tile.farm:crimson_roots_farm.name', maxLevel: 2, description: 'farm.info.desc.nether_roots', requireGamerules: [  ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_bone_meal', name: 'tile.farm:bone_meal_farm.name', maxLevel: 4, description: 'farm.info.desc.bone_meal', requireGamerules: [  ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.none' },
	{ buttonTexture: 'button_tree_farm', name: 'tile.farm:auto_tree_farm.name', maxLevel: 1, description: 'farm.info.desc.auto_tree_farm', requireGamerules: [ 'mobgriefing' ], dimensionRestriction: 'farm.info.dimension.overworld_end', biomeRestriction: 'farm.info.biome.no_warm_biomes' },
	{ buttonTexture: 'button_tree_farm', name: 'tile.farm:tree_farm.name', maxLevel: 1, description: 'farm.info.desc.tree_farm', requireGamerules: [  ], dimensionRestriction: 'farm.info.dimension.none', biomeRestriction: 'farm.info.biome.none' },
]
function FarmInfoUI (player) {
	const form = new ActionFormData()
		.title({ translate: 'farm.title.addon_info_farms' })
		.body({ translate: 'farm.body.farm_info', with: [ '\n' ] })
		.button({ translate: 'farm.button.material_categories' });
	farmInfoList.forEach(farm => {
		form.button({ translate: farm.name }, `textures/farm/button/${farm.buttonTexture}`);
	});
	form.button({ translate: 'farm.button.back' }, "textures/farm/button/back")
	.show(player).then(r => {
		if (r.canceled) return;
		const selectionIndex = r.selection;
        if (selectionIndex === 0) {
            AllowedMaterialsUI(player);
        } else if (selectionIndex === farmInfoList.length + 1) {
            AddonInfoUI(player);
        } else {
            ShowFarmInfo(player, farmInfoList[selectionIndex - 1])
        }
	});
}
function ShowFarmInfo (player, selectedFarm) {
	const rawBody = { rawtext: [
		{ translate: 'farm.info.max_level', with: [ `${selectedFarm.maxLevel}`, '\n' ] },
		{ translate: 'farm.info.description' },
		{ translate: selectedFarm.description, with: [ '\n' ] },
		{ translate: 'farm.info.gamerules', with: [ '\n' ] }
	] }
	if (selectedFarm.requireGamerules.length > 0) {
        selectedFarm.requireGamerules.forEach(gamerule => {
            rawBody.rawtext.push({ text: `§r§7§o• ${gamerule}\n` });
        });
    } else {
        rawBody.rawtext.push({ translate: 'farm.info.no_gamerules', with: ['\n'] });
    }
	rawBody.rawtext.push(
		{ translate: 'farm.info.dimension_restriction' },
		{ translate: selectedFarm.dimensionRestriction, with: [ '\n' ] },
		{ translate: 'farm.info.biome_restriction' },
		{ translate: selectedFarm.biomeRestriction, with: [ '\n' ] }
	)
	new ActionFormData()
	.title({ translate: selectedFarm.name })
	.body(rawBody)
	.button({ translate: 'farm.button.back' }, "textures/farm/button/back")
	.show(player).then(r => {
		if (r.canceled) return;
		[FarmInfoUI][r.selection](player);
	})
}
function AllowedMaterialsUI (player) {
	new ActionFormData()
	.title({ translate: 'farm.title.material_categories' })
	.body({ translate: 'farm.body.material_categories', with: [ '\n' ] })
	.button({ translate: 'farm.button.back' }, "textures/farm/button/back")
	.show(player).then(r => {
		if (r.canceled) return;
		[AddonInfoUI][r.selection](player);
	});
}
function RedstoneTableInfoUI (player) {
	new ActionFormData()
	.title({ translate: 'farm.title.redstone_table' })
	.body({ translate: 'farm.body.redstone_table', with: [ '\n' ] })
	.button({ translate: 'farm.button.back' }, "textures/farm/button/back")
	.show(player).then(r => {
		if (r.canceled) return;
		[AddonInfoUI][r.selection](player);
	});
}
function ItemsInfoUI (player) {
	new ActionFormData()
	.title({ translate: 'farm.title.addon_info_items' })
	.button({ translate: 'farm.button.farm_booster' }, "textures/farm/items/farm_booster")
	.button({ translate: 'farm.button.cardboard_box' }, "textures/farm/items/cardboard_box")
	.button({ translate: 'farm.button.back' }, "textures/farm/button/back")
	.show(player).then(r => {
		if (r.canceled) return;
		[FarmBoosterInfoUI, BoxInfoUI, AddonInfoUI][r.selection](player);
	});
}
function FarmBoosterInfoUI (player) {
	new ActionFormData()
	.title({ translate: 'farm.title.farm_booster' })
	.body({ translate: 'farm.body.farm_booster', with: [ '\n' ] })
	.button({ translate: 'farm.button.back' }, "textures/farm/button/back")
	.show(player).then(r => {
		if (r.canceled) return;
		[ItemsInfoUI][r.selection](player);
	});
}
function BoxInfoUI (player) {
	new ActionFormData()
	.title({ translate: 'farm.title.cardboard_box' })
	.body({ translate: 'farm.body.box_info', with: [ '\n' ] })
	.button({ translate: 'farm.button.back' }, "textures/farm/button/back")
	.show(player).then(r => {
		if (r.canceled) return;
		[ItemsInfoUI][r.selection](player);
	});
}
function ConfigUI (player) {
	new ModalFormData()
	.title({ translate: 'farm.title.insta_farm_settings' })
	.slider({ translate: 'farm.slider.material_percentage' }, 0, 100, 1, InstaFarmConfig.blockPercentage)
	.slider({ translate: 'farm.slider.sound_volume' }, 0, 100, 1, InstaFarmConfig.volume)
	.toggle({ translate: 'farm.toggle.generate_particles' }, InstaFarmConfig.showParticles)
	.show(player).then(r => {
		if (r.canceled) return;
		const [ percentage, volume, showParticles ] = r.formValues;
		const configRawtext = { rawtext: [] }
		if (InstaFarmConfig.blockPercentage !== percentage) {
			configRawtext.rawtext.push({ translate: 'farm.message.update_config_percentage', with: [ '\n', `${percentage}` ] });
		}
		if (InstaFarmConfig.volume !== volume) {
			configRawtext.rawtext.push({ translate: 'farm.message.update_config_volume', with: [ '\n', `${volume}` ] });
		}
		if (InstaFarmConfig.showParticles !== showParticles) {
			configRawtext.rawtext.push({ translate: 'farm.message.update_config_particles', with: [ '\n', `${showParticles}` ] });
		}
		if (configRawtext.rawtext.length > 0) {
			configRawtext.rawtext.unshift({ translate: 'farm.message.update_config_start' })
			player.dimension.playSound('if.approve_sound', player.location, { volume: volume / 100 });
			world.sendMessage(configRawtext);
			updateConfig({ blockPercentage: percentage, showParticles: showParticles, volume: volume });
		}
	});
}
//--------
function updateConfig(newValues) {
    Object.assign(InstaFarmConfig, newValues);
    world.setDynamicProperty("InstaFarmConfig", JSON.stringify(InstaFarmConfig));
}
function getDefaultMaterial(material) {
    const defaultMaterials = {
        leaves: "oak_leaves",
        general_coral_fan: "brain_coral_fan",
        planks: "oak_planks",
        general_terracotta: "hardened_clay",
		solid_block: "cobblestone",
		wool: "white_wool"
    };
    return defaultMaterials[material] || material;
}
system.afterEvents.scriptEventReceive.subscribe(e => {
	const { id, sourceEntity } = e;
	if (!sourceEntity) return;
	if (sourceEntity.typeId !== "farm:material_list") return;
	if (sourceEntity.getProperty("farm:wait_upgrade") === true) return;
	if (id === "farm:create_boxes") {
		const materialsData = sourceEntity.getDynamicProperty("materials");
		if (!materialsData) return;
		const materials = JSON.parse(materialsData);
		for (const [material, value] of Object.entries(materials)) {
			const [current, total] = value.split("/").map(Number);
			if (current === 0) continue;
			const tempItem = new ItemStack(`minecraft:${getDefaultMaterial(material)}`, 1)
			sourceEntity.dimension.spawnItem(current > tempItem.maxAmount ? createBox({ materialId: `minecraft:${getDefaultMaterial(material)}`, amount: current }) : new ItemStack(`minecraft:${getDefaultMaterial(material)}`, current), sourceEntity.location);
		}
		sourceEntity.remove();
	}
});
function createBox(material) {
	const box = new ItemStack("farm:cardboard_box", 1);
	box.setDynamicProperty("farm:material_info", JSON.stringify(material));
	box.setLore([`§o§7${material.materialId.replace("minecraft:", "").split('_').map(word => word.charAt(0).toUpperCase() + word.slice(1)).join(' ')} x${material.amount}`]);
	return box;
}
class IfCardboardBox {
	onUse(e) {
		const { itemStack, source } = e;
		const material = itemStack.getDynamicProperty("farm:material_info");
		if (material) {
			const materialObj = JSON.parse(material);
			const tempItem = new ItemStack(materialObj.materialId, 1);
			const maxAmount = tempItem.maxAmount;
    		const maxRetirar = Math.min(materialObj.amount, (maxAmount * 16));
			const rawtext = {
				rawtext: [
					{ translate: 'farm.slider.remove_items_from_box1', with: [ '\n' ] },
					{ translate: `farm.${materialObj.materialId}.t` },
					{ text: ` x${materialObj.amount}\n` }
				]
			}
			if (materialObj.amount > maxAmount * 16) rawtext.rawtext.push({ translate: 'farm.slider.remove_items_from_box2', with: [ '\n',`${maxAmount * 16}` ] });
			rawtext.rawtext.push({ translate: 'farm.slider.remove_items_from_box3', with: [ '\n' ] })
			new ModalFormData()
				.title({ translate: 'farm.title.remove_items_from_box' })
				.slider(rawtext, 1, maxRetirar, 1, 1)
				.show(source)
				.then(r => {
					if (r.canceled) return;
					const [value] = r.formValues;
					const newAmount = materialObj.amount - value;
					const spawnItem = (amount) => source.dimension.spawnItem(new ItemStack(materialObj.materialId, amount), source.location);
					Array(Math.floor(value / maxAmount)).fill().forEach(() => spawnItem(maxAmount));
            		if (value % maxAmount > 0) spawnItem(value % maxAmount);
					if (newAmount > 0) {
						source.getComponent("inventory").container.setItem(source.selectedSlotIndex, createBox({
							materialId: materialObj.materialId,
							amount: newAmount
						}));
					} else {
						decrementItemInHand(source, true);
					}
				});
		}
	}
}
class IfStartConfig {
	onUse(e) {
		AddonInfoUI(e.source);
	}
}
class IfSetFarm {
	onPlace(e) {
		checkBiomeCondition(farmMap.get(selectFarmMap.get(e.block.typeId)[0].farmId), e.block);
	}
	onPlayerInteract(e) {
		const { player, block, dimension } = e;
		const testFarmMaterials = dimension.getEntitiesAtBlockLocation(block.location).find(entity => entity.getComponent("minecraft:type_family").hasTypeFamily("farm_blocks"));
		if (!testFarmMaterials) {
			const farmList = selectFarmMap.get(block.typeId);
			const selectFarmUI = new ActionFormData()
				.title({ translate: 'farm.title.select_farm' });
			farmList.forEach(farm => { selectFarmUI.button(farm.farmName, `textures/farm/button/${farm.button}`) })
			
			selectFarmUI.show(player).then(r => {
				if (r.canceled) return;
				setupInitialMaterials(block, block.permutation, dimension, farmList[r.selection].farmId);
			})
		} else {
			manageFarmMaterials(player, block, dimension, testFarmMaterials.getDynamicProperty("farmId"));
		}
	}
}
class IfPlaceBlock {
	onPlace(e) {
		setupInitialMaterials(e.block, e.block.permutation, e.dimension, e.block.typeId);
		checkBiomeCondition(farmMap.get(e.block.typeId), e.block);
	}
	onPlayerInteract(e) {
		manageFarmMaterials(e.player, e.block, e.dimension, e.block.typeId);
	}
}
function reorderMaterials(materials) {
	const entries = Object.entries(materials);
	return Object.fromEntries(
		entries.sort(([_, progressA], [__, progressB]) => {
			const [currentA, maxA] = progressA.split("/").map(Number);
			const [currentB, maxB] = progressB.split("/").map(Number);
			return currentA >= maxA && currentB < maxB ? 1 : currentA < maxA && currentB >= maxB ? -1 : 0;
		})
	);
}
function FarmSetMaterials(blockId, entity, level) {
	const percentage = InstaFarmConfig.blockPercentage;
	const farmData = farmMap.get(blockId);
	if (!farmData) return;
	const { maxLevel, levels } = farmData;
	const farmConfig = (maxLevel && level > 0 && level <= maxLevel.level) ? levels[maxLevel.baseLevel] : levels[level];
	if (!farmConfig) return;
	const materialsProgress = {};
	for (let material in farmConfig.materials) {
		const materialData = farmConfig.materials[material];
		const adjustedAmount = Math.round((materialData * percentage) / 100);
		if (adjustedAmount > 0) materialsProgress[material] = `0/${adjustedAmount}`;
	}
	entity.setDynamicProperty("materials", JSON.stringify(materialsProgress));
}
//---------------CONSUMIR MATERIALES INVENTARIO---------------//
function updateMaterials(player, materials, entity) {
	const isCreativeMode = isCreative(player);
	const inventory = player.getComponent("inventory").container;
	const materialsToConsume = [];
	for (const [material, progress] of Object.entries(materials)) {
		let [current, max] = progress.split("/").map(Number);
		let required = max - current;
		if (required <= 0) continue;
		if (!isCreativeMode) {
			for (let slot = 0; slot < inventory.size && required > 0; slot++) {
				const item = inventory.getItem(slot);
				if (!item) continue;
				const itemId = item.typeId.replace("minecraft:", "");
				const isValidItem = allowedMaterialList[material] ? allowedMaterialList[material].includes(itemId) : itemId === material;
				if (isValidItem) {
					const available = item.amount;
					if (available > required) {
						current += required;
						item.amount -= required;
						materialsToConsume.push({ slot: slot, item: item, amount: required });
						required = 0;
					} else {
						current += available;
						required -= available;
						materialsToConsume.push({ slot: slot, item: undefined, id: item.typeId, amount: item.amount });
					}
					
				}
			}
		} else {
			current = max;
		}
		materials[material] = `${current}/${max}`;
	}
	if (materialsToConsume.length > 0 || isCreativeMode) {
		const updateMaterials = () => {
			entity.setDynamicProperty("materials", JSON.stringify(materials));
			player.sendMessage({ translate: 'farm.message.materials_updated', with: [ '\n' ] });
			player.dimension.playSound('if.approve_sound', player.location, { volume: InstaFarmConfig.volume / 100 });
		};
		if (isCreativeMode) {
			updateMaterials();
		} else {
			const message = createMaterialUsageMessage(materialsToConsume);
			new ActionFormData()
				.title({ translate: 'farm.title.confirm_materials_to_consume' })
				.body(message)
				.button({ translate: 'farm.button.accept' }, 'textures/farm/button/accept')
				.button({ translate: 'farm.button.cancel' }, 'textures/farm/button/cancel')
				.show(player)
				.then(r => {
					if (r.selection === 0) {
						materialsToConsume.forEach(material => {
							inventory.setItem(material.slot, material.item);
						});
						updateMaterials();
					} else {
						player.dimension.playSound("if.deny_sound", player.location, { volume: InstaFarmConfig.volume / 100 });
						player.sendMessage({ translate: 'farm.message.cancel_materials_to_consume' });
					}
				});
		}
	} else {
		player.dimension.playSound("if.deny_sound", player.location, { volume: InstaFarmConfig.volume / 100 });
		player.sendMessage({ translate: 'farm.message.no_materials_in_inventory' });
	}
}
function createMaterialUsageMessage(materialsToConsume) {
	const message = { rawtext: [] };
	message.rawtext.push({ translate: 'farm.body.confirm_materials_to_consume' });
	materialsToConsume.forEach(material => {
		message.rawtext.push(
			{ text: `§7- ` },
			{ translate: `farm.${material.item ? material.item.typeId : material.id}.t` },
			{ text: `§3 x${material.amount} §7(Slot ${material.slot})\n` }
		);
	});
	return message;
}
//------------------------------------------------------------//
function setupInitialMaterials(block, permutation, dimension, farmId) {
	const entity = dimension.spawnEntity("farm:material_list", block.center());
	entity.triggerEvent(`${farmId}`);
	FarmSetMaterials(farmId, entity, 0);
	entity.setDynamicProperty("farmId", farmId);
	entity.setRotation({ x: 0, y: { north: 180, south: 0, west: 90, east: -90 }[permutation.getState("minecraft:cardinal_direction")] });
}
//----------- INTERACCION JUGADOR -------------//
function isValidStructurePlacement(block, direction, selectedFarmLevel) {
	const {  height, yOffset = 0 } = selectedFarmLevel.dimensions ? selectedFarmLevel.dimensions[direction] : selectedFarmLevel;
	const heightRange = block.dimension.heightRange;
    const minWorldHeight = heightRange.min;
    const maxWorldHeight = heightRange.max;
    return ((block.y + yOffset) >= minWorldHeight) && ((block.y + height + yOffset) <= maxWorldHeight);
}
function manageFarmMaterials(player, block, dimension, farmId) {
    const direction = block.permutation.getState("minecraft:cardinal_direction");
    const selectedFarm = farmMap.get(farmId);
    const inventory = player.getComponent("inventory").container;
    const item = inventory.getItem(player.selectedSlotIndex);
    const entity = dimension.getEntitiesAtBlockLocation(block.location).find(entity => entity.getComponent("minecraft:type_family").hasTypeFamily("farm_blocks"));
    if (selectedFarm?.additionalConditions?.(block) === false) return;
    if (!entity) {
		player.sendMessage({ translate: 'farm.message.this_farm_has_no_data' });
		player.dimension.playSound('if.deny_sound', player.location, { volume: InstaFarmConfig.volume / 100 });
		return;
	}
    const level = entity.getProperty("farm:level");
    const selectedFarmLevel = getSelectedFarmLevel(selectedFarm, level);
    if (!isValidStructurePlacement(block, direction, selectedFarmLevel)) {
		player.dimension.playSound('if.deny_sound', player.location, { volume: InstaFarmConfig.volume / 100 });
		player.sendMessage({ translate: 'farm.message.generate_outside_world', with: [ '\n' ] });
        return;
    }
    if (entity.getProperty("farm:wait_upgrade")) {
        handleUpgrade(player, block, entity, selectedFarm, item, farmId, level, direction);
        return;
    }
    handleMaterialManagement(player, entity, block, direction, selectedFarmLevel);
}
function getSelectedFarmLevel(selectedFarm, level) {
    const { maxLevel, levels } = selectedFarm;
    const baseLevel = maxLevel?.baseLevel ?? 0;
    const yOffsetIncrement = maxLevel?.yOffset ?? 0;
    return levels[level] ?? { ...levels[baseLevel], yOffset: levels[baseLevel].yOffset + (level - baseLevel) * yOffsetIncrement };
}
function handleUpgrade(player, block, entity, selectedFarm, item, farmId, level, direction) {
    const { levels, maxLevel } = selectedFarm;
    const highestLevel = maxLevel?.level ?? Math.max(...Object.keys(levels).map(Number));
    const baseLevel = maxLevel?.baseLevel ?? 0;
    const yOffsetIncrement = maxLevel?.yOffset ?? 0;
    const nextLevel = level + 1;
    if (nextLevel > highestLevel) {
		player.dimension.playSound("if.deny_sound", block.center(), { volume: InstaFarmConfig.volume / 100 });
		player.sendMessage({ translate: 'farm.message.no_more_upgrades', with: [ '\n' ] });
		return;
	}
    const nextFarmLevel = levels[nextLevel] ?? { ...levels[baseLevel], yOffset: levels[baseLevel].yOffset + (nextLevel - baseLevel) * yOffsetIncrement };
    if (!isValidStructurePlacement(block, direction, nextFarmLevel)) {
		player.dimension.playSound("if.deny_sound", block.center(), { volume: InstaFarmConfig.volume / 100 });
		player.sendMessage({ translate: 'farm.message.next_level_limit', with: ["\n"] });
        return;
    }
    if (item?.typeId === "farm:farm_booster") {
		player.dimension.playSound("if.approve_sound", block.center(), { volume: InstaFarmConfig.volume / 100 });
		player.sendMessage({ translate: 'farm.message.farm_upgraded', with: [ '\n', `${nextLevel + 1}` ] })
        entity.setProperty("farm:level", nextLevel);
        entity.setProperty("farm:wait_upgrade", false);
        FarmSetMaterials(farmId, entity, nextLevel);
        decrementItemInHand(player);
    } else {
        player.sendMessage({ translate: 'farm.message.use_farm_booster', with: [ '\n' ] });
		player.dimension.playSound("if.deny_sound", block.center(), { volume: InstaFarmConfig.volume / 100 });
    }
}
function handleMaterialManagement(player, entity, block, direction, selectedFarmLevel) {
    let materials = JSON.parse(entity.getDynamicProperty("materials"));
    materials = reorderMaterials(materials);
    const allMaterialsComplete = Object.values(materials).every(progress => {
        const [current, max] = progress.split("/").map(Number);
        return current >= max;
    });
    const materialMessage = generateMaterialMessage(materials, allMaterialsComplete);
    const farmUI = new ActionFormData()
        .title({ translate: 'farm.title.required_materials' })
        .body(materialMessage);
    farmUI.button(allMaterialsComplete ? { translate: 'farm.button.generate_structure' } : { translate: 'farm.button.add_materials' }, allMaterialsComplete ? 'textures/farm/button/build' : 'textures/farm/button/insert');
    farmUI.show(player).then(r => {
        if (r.canceled) return;
        if (allMaterialsComplete) {
            getStructurePlacement(block, direction, selectedFarmLevel, player);
            entity.setProperty("farm:wait_upgrade", true);
        } else {
            updateMaterials(player, materials, entity);
        }
    });
}
function generateMaterialMessage(materials, allMaterialsComplete) {
    const materialMessage = { rawtext: [] };
    if (allMaterialsComplete) materialMessage.rawtext.push({ translate: 'farm.message.all_materials_complete', with: [ '\n' ] });
    for (const [material, progress] of Object.entries(materials)) {
        const [current, max] = progress.split("/").map(Number);
        materialMessage.rawtext.push(
            { text: `§f` },
            { translate: `farm.minecraft:${material}.t` },
            { text: `: ${current < max ? "§c" : "§a"}${progress}\n` }
        );
    }
    return materialMessage;
}
//---------------------------------------------//
function checkBiomeCondition(farmConfig, block) {
	if (farmConfig?.event) {
		const detector = block.dimension.spawnEntity("farm:biome_detector", block.center());
		detector.triggerEvent(farmConfig.event);
	}
	if (farmConfig?.onPlaceMessage) {
		block.dimension.getPlayers({ closest: 1, location: block.location })[0].sendMessage({ translate: farmConfig.onPlaceMessage, with: [ '\n' ] });
	}
}
function getStructurePlacement(block, direction, farmBlock, player) {
	const { width, height, depth, offset = 0, depthOffset = 0, yOffset = 0, summonEntities, fixStructure, fixBlocks, addItemToBlock, structures, teleportPlayer } = farmBlock.dimensions?.[direction] ?? farmBlock;
	const buildAnimation = true;
	const isWidthEven = (width % 2 === 0);
	const dimension = block.dimension;
	let pos = block.location;
	const structureData = typeof structures === 'object' ? structures[direction] : null;
	const structureId = structureData?.id ?? structures;
	const structureRotation = structureData?.rotation ?? {
		north: "None",
		south: "Rotate180",
		east: "Rotate90",
		west: "Rotate270"
	}[direction];
	const adjustments = {
		north: { x: -offset - Math.floor(width / 2), z: 1 + depthOffset },
		south: { x: offset - Math.floor(width / 2) + (isWidthEven ? 1 : 0), z: -depth - depthOffset },
		east: { x: -depth - depthOffset, z: -offset - Math.floor(width / 2) },
		west: { x: 1 + depthOffset, z: offset - Math.floor(width / 2) + (isWidthEven ? 1 : 0) }
	};
	pos = {
		x: pos.x + adjustments[direction].x,
		y: pos.y + yOffset,
		z: pos.z + adjustments[direction].z
	};
	world.structureManager.place(
		structureId,
		dimension,
		pos,
		{
			animationMode: "Layers",
			animationSeconds: height * 0.1,
			rotation: structureRotation
		}
	);
	dimension.playSound("if.build_sound", block.center(), { volume: InstaFarmConfig.volume / 100 });
	if (InstaFarmConfig.showParticles) {
		const particle = new MolangVariableMap();
		const isNorthSouth = direction === 'north' || direction === 'south';
		const sizeX = (isNorthSouth ? width : depth) * 0.5 + 1;
		const sizeZ = (isNorthSouth ? depth : width) * 0.5 + 1;
		const particleLocation = { x: pos.x + (isNorthSouth ? width : depth) / 2, y: pos.y, z: pos.z + (isNorthSouth ? depth : width) / 2 };
		particle.setFloat('variable.size_x', sizeX);
		particle.setFloat('variable.size_z', sizeZ);
		particle.setFloat('variable.size_y', height);
		dimension.spawnParticle("farm:build_particle", particleLocation, particle);
	}
	if (summonEntities) {
		const { type, positions } = summonEntities;
		const offsets = {
			north: (offset, depthOffset) => ({ x: block.x + offset + 0.5, z: block.z + depthOffset + 0.5 }),
			south: (offset, depthOffset) => ({ x: block.x - offset + 0.5, z: block.z - depthOffset + 0.5 }),
			east:  (offset, depthOffset) => ({ x: block.x - depthOffset + 0.5, z: block.z + offset + 0.5 }),
			west:  (offset, depthOffset) => ({ x: block.x + depthOffset + 0.5, z: block.z - offset + 0.5 })
		};
		system.runTimeout(() => {
			positions.forEach(({ offset, yOffset, depthOffset }) => {
				const { x, z } = offsets[direction](offset, depthOffset);
				const y = block.y + yOffset;
				dimension.runCommandAsync(`summon ${type} ${x} ${y} ${z} ~~`);
			});
		}, height * 2);
	}	
	if (fixStructure && buildAnimation) {
		system.runTimeout(() => {
			world.structureManager.place(
				fixStructure,
				dimension,
				pos,
				{
					rotation: structureRotation
				}
			);
		}, (height * 2));
	}
	if (fixBlocks) {
		system.runTimeout(() => {
			fixBlocks.forEach(blockData => {
				blockData.offsets.forEach(offset => {
					const blockOffset = getBlockOffset(offset, direction);
					const targetBlock = block.offset(blockOffset);
					if (blockData.state) {
						const stateName = blockData.state.name;
						const stateValue = blockData.state[direction];
						targetBlock.setPermutation(BlockPermutation.resolve(blockData.type, { [stateName]: stateValue }));
					} else targetBlock.setPermutation(BlockPermutation.resolve(blockData.type));
				});
			});
		}, height * 2);
	}
	if (addItemToBlock) {
		system.runTimeout(() => {
			addItemToBlock.forEach(({ id, quantity, offset }) => {
				let blockOffset = getBlockOffset(offset, direction);
				const targetInventory = block.offset(blockOffset).getComponent("inventory")?.container;
				if (!targetInventory) return;
				targetInventory.addItem(new ItemStack(id, quantity));
			});
		}, height * 2);
	}
	if (teleportPlayer) {
		let blockOffset = getBlockOffset(teleportPlayer, direction);
		player.teleport(block.offset(blockOffset).location);
	}
	/*
	backup
	const to = {
		x: isNorthSouth ? pos.x + width : pos.x + depth,
		y: pos.y + height,
		z: isNorthSouth ? pos.z + depth : pos.z + width
	};
	const structureBackUp = `${structureId}_${level}_${block.x}_${block.y}_${block.z}`;
	world.structureManager.createFromWorld(
		structureBackUp,
		block.dimension,
		pos,
		to,
		{ saveMode: StructureSaveMode.World, includeEntities: false }
	);*/
}
function getBlockOffset(offset, direction) {
	let { x, y, z } = offset;
	switch (direction) {
		case "north": return { x, y, z };
		case "south": return { x: -x, y, z: -z };
		case "east": return { x: -z, y, z: x };
		case "west": return { x: z, y, z: -x };
	}
}